#!/usr/bin/local/python3.9
# -*- coding: utf-8 -*-

"""
    Builds CLI options
"""

from argparse import ArgumentParser
from os import getcwd
from os import makedirs
from os.path import isdir
from os.path import join
from os.path import sep
from pathlib import Path

from pdf_play.__settings__ import COLORS
from pdf_play.__settings__ import FONTS
from pdf_play.__settings__ import FONT_SIZES
from pdf_play.helpers import _input_actions as actions
from pdf_play.helpers import _input_types as types

_pdf_play = f'''\n{"-" * 100}\n{"*" * 30}{" " * 16}PDF-Play{" " * 16}{"*" * 30
}\n{"-" * 100}\n'''


def _validate_otm(args):
    if args.output_directory is None:
        args.output_directory = join(getcwd(), 'watermarked')
        if not isdir(args.output_directory):
            # print('User did not provide an output location, a directory called '
            #       '"watermarked" will be created to store all the watermarked PDF(s)')
            makedirs(args.output_directory)
    print(f'Watermarked files will be saved to: {args.output_directory}')


def _validate_mto(args):
    _validate_otm(args)
    texts = []
    for item in args.texts:
        if isinstance(item, str):
            name, text = item, item
        else:
            name, text = item
        texts.append([name, text])
    args.texts = texts


def _validate_oto(args):
    if args.output_file is None:
        base = Path(args.target_file).parent.resolve()
        file_name = args.target_file.split(sep)[-1]
        name, extension = file_name.split('.')
        args.output_file = join(base, f'{name}_watermarked.{extension}')
    print(f'Watermarked files will be saved to: {args.output_file}')


def _update_args(args):
    if args.type == 'oto':
        _validate_oto(args)
    if args.type == 'otm':
        _validate_otm(args)
    if args.type == 'mto':
        _validate_mto(args)
    return args


def _validate_args(parser, args):
    if args.command is None:
        print(_pdf_play)
        parser.print_help()
        exit(0)
    else:
        if args.type is None:
            print('\n--> User must specify the mode of operation. '
                  f'Get help: {args.command} -h')
            exit(0)


def parse_user_args(command=None):
    __usage = 'usage: pdf_play [-h] {watermark} ..'
    _usage = f'{_pdf_play}\n{__usage}'
    try:
        parser = ArgumentParser(prog='pdf_play',
                                description='A Python utility to watermark PDF '
                                            'documents',
                                epilog="""--> Run -h/--help against indivdual commands to 
                                get more details. Example: watermark -h""")
        parser.add_argument('-v', '--version', action='version',
                            version='PDF-Play 1.2.1',
                            help='display version of the program')
        parser.add_argument('-d', '--debug', action='store_true',
                            default=False, dest='debug',
                            help='display stack trace for a failure')

        commands = parser.add_subparsers(dest='command',
                                         help="""Commands available in PDF-Play, 
                                         currently there's just one :(""")
        watermark = commands.add_parser('watermark', help='Watermarks PDF files',
                                        description='Apply watermarks to PDF '
                                                    'files',
                                        epilog="""--> Run -h/--help against positional 
                                        arguments to get more details. Example: watermark 
                                        oto -h""")
        sub_commands = watermark.add_subparsers(dest='type',
                                                help="""Modes of operation for 
                                                     watermarking PDF files""")
        oto = sub_commands.add_parser('oto',
                                      help="""One-To-One: Apply watermark text to a 
                                           single file""",
                                      description="""One-To-One: Use this mode when you 
                                      have got a single target file""",
                                      epilog="""--> Sample Usage: watermark oto -t 'this 
                                      is my watermark text' -i 
                                      'MyPC/Downloads/PDFFiles/sample.pdf'""")
        oto.add_argument('-t', '--text', type=str, default='PDFPlay',
                         dest='text', required=True, nargs='+',
                         action=actions.watermark_text,
                         help='(**) Text that is to be applied as the watermark')
        oto.add_argument('-i', '--input', default=None, type=types.target_file_oto,
                         action=actions.target_file_oto, dest='target_file',
                         required=True,
                         help='(**) Path to the PDF file that is to be watermarked')
        oto.add_argument('-o', '--output', default=None, type=types.output_file_oto,
                         action=actions.output_file_oto, dest='output_file',
                         help='Name of the output file, by default '
                              '"{input-file}_watermarked.pdf" will be generated')
        oto.add_argument('-fn', '--font-name', default='Helvetica-Bold', type=str,
                         choices=FONTS, dest='font_name',
                         help='Name of the font that you want to use in the watermark')
        oto.add_argument('-fs', '--font-size', default='medium', type=str,
                         choices=FONT_SIZES, dest='font_size',
                         help='Size of the font')
        oto.add_argument('-fc', '--font-color', default='lightred', type=str,
                         choices=COLORS, dest='font_color',
                         help='Color of the font')
        oto.add_argument('-ta', '--text-alignment', default='diagonal', type=str,
                         choices=['horizontal', 'diagonal'], dest='text_alignment',
                         help='Alignment of the watermark in the document')
        oto.add_argument('-p', '--print', action='store_true',
                         default=False, dest='verbose',
                         help='Display informational messages')

        otm = sub_commands.add_parser('otm', help='One-To-Many: Apply watermark to '
                                                  'multiple files',
                                      description='One-To-Many: Use this mode when you '
                                                  'have got multiple target/S files',
                                      epilog=""""--> Sample Usage: watermark otm -t 'this 
                                      is my watermark text' -i 'sample1.pdf' 'sample2.pdf' 
                                      'MyPC/Downloads/PDFFiles'""")
        otm.add_argument('-t', '--text', type=str, default='PDFPlay', dest='text',
                         required=True, action=actions.watermark_text, nargs='+',
                         help='(**) Text that is to be applied as the watermark')
        otm.add_argument('-i', '--input', default=None, type=types.target_file_otm,
                         action=actions.target_file_otm, dest='target_files', nargs='+',
                         required=True,
                         help="""(**) Paths to files/directories that contain
                         your target files""")
        otm.add_argument('-o', '--output', default=None,
                         type=types.output_file_otm,
                         action=actions.output_file_otm, dest='output_directory',
                         help='Directory to save watermarked files')
        otm.add_argument('-fn', '--font-name', default='Helvetica-Bold', type=str,
                         choices=FONTS, dest='font_name',
                         help='Name of the font that you want to use in the watermark')
        otm.add_argument('-fs', '--font-size', default='medium', type=str,
                         choices=FONT_SIZES, dest='font_size',
                         help='Size of the font')
        otm.add_argument('-fc', '--font-color', default='lightred', type=str,
                         choices=COLORS, dest='font_color',
                         help='Color of the font')
        otm.add_argument('-ta', '--text-alignment', default='diagonal', type=str,
                         choices=['horizontal', 'diagonal'], dest='text_alignment',
                         help='Alignment of the watermark in the document')
        otm.add_argument('-p', '--print', action='store_true',
                         default=False, dest='verbose',
                         help='Display informational messages')

        mto = sub_commands.add_parser('mto', help='Many-To-One: Apply many different '
                                                  'watermarks to a single file',
                                      description="""Many-To-One: Use this mode when you
                                      want to watermark a file multiple times with
                                      different watermark texts""",
                                      epilog="""--> Sample usage: watermark mto -t 
                                      'watermark text#1' 'watermark text#2' 
                                      'MyPC/SampleWatermarks/sample1.txt' -i 'sample.pdf'
                                      """)
        mto.add_argument('-t', '--text', default='PDFPlay',
                         type=types.watermark_mto,
                         action=actions.watermark_mto,
                         dest='texts', nargs='+',
                         help="""(**) Texts that are to be watermarked or path to a txt
                              file that contains such texts""")
        mto.add_argument('-i', '--input', default=None, type=types.target_file_mto,
                         action=actions.target_file_oto, dest='target_file',
                         required=True,
                         help='(**) Path to the PDF file that is to be watermarked')
        mto.add_argument('-o', '--output', default=None,
                         type=types.output_file_otm,
                         action=actions.output_file_otm, dest='output_directory',
                         help='Directory to save watermarked files')
        mto.add_argument('-fn', '--font-name', default='Helvetica-Bold', type=str,
                         choices=FONTS, dest='font_name',
                         help='Name of the font that you want to use in the watermark')
        mto.add_argument('-fs', '--font-size', default='medium', type=str,
                         choices=FONT_SIZES, dest='font_size',
                         help='Size of the font')
        mto.add_argument('-fc', '--font-color', default='lightred', type=str,
                         choices=COLORS, dest='font_color',
                         help='Color of the font')
        mto.add_argument('-ta', '--text-alignment', default='diagonal', type=str,
                         choices=['horizontal', 'diagonal'], dest='text_alignment',
                         help='Alignment of the watermark in the document')
        mto.add_argument('-p', '--print', action='store_true',
                         default=False, dest='verbose',
                         help='Display informational messages')
        args = parser.parse_args() if command is None else parser.parse_args(command)
        _validate_args(parser, args)
        args = _update_args(args)
        return args
    except Exception as e:
        print(f'Error --> Bad user-input: {e}')
        raise e


def parse_watermark_args():
    try:
        parser = ArgumentParser(prog='watermark',
                                description='A Python utility to watermark PDF '
                                            'documents',
                                epilog="""--> Run -h/--help against positional 
                                        arguments to get more details. Example: watermark 
                                        oto -h""")
        sub_commands = parser.add_subparsers(dest='type',
                                             help="""Modes of operation for 
                                                 watermarking PDF files""")
        oto = sub_commands.add_parser('oto',
                                      help="""One-To-One: Apply watermark text to a 
                                                   single file""",
                                      description="""One-To-One: Use this mode when you 
                                              have got a single target file""",
                                      epilog="""--> Sample Usage: watermark oto -t 'this 
                                              is my watermark text' -i 
                                              'MyPC/Downloads/PDFFiles/sample.pdf'""")
        oto.add_argument('-t', '--text', type=str, default='PDFPlay',
                         dest='text', required=True, nargs='+',
                         action=actions.watermark_text,
                         help='(**) Text that is to be applied as the watermark')
        oto.add_argument('-i', '--input', default=None, type=types.target_file_oto,
                         action=actions.target_file_oto, dest='target_file',
                         required=True,
                         help='(**) Path to the PDF file that is to be '
                              'watermarked')
        oto.add_argument('-o', '--output', default=None, type=types.output_file_oto,
                         action=actions.output_file_oto, dest='output_file',
                         help='Name of the output file, by default '
                              '"{input-file}_watermarked.pdf" will be generated')
        oto.add_argument('-fn', '--font-name', default='Helvetica-Bold', type=str,
                         choices=FONTS, dest='font_name',
                         help='Name of the font that you want to use in the watermark')
        oto.add_argument('-fs', '--font-size', default='medium', type=str,
                         choices=FONT_SIZES, dest='font_size',
                         help='Size of the font')
        oto.add_argument('-fc', '--font-color', default='lightred', type=str,
                         choices=COLORS, dest='font_color',
                         help='Color of the font')
        oto.add_argument('-ta', '--text-alignment', default='diagonal', type=str,
                         choices=['horizontal', 'diagonal'], dest='text_alignment',
                         help='Alignment of the watermark in the document')
        oto.add_argument('-p', '--print', action='store_true',
                         default=False, dest='verbose',
                         help='Display informational messages')

        otm = sub_commands.add_parser('otm', help='One-To-Many: Apply watermark to '
                                                  'multiple files',
                                      description='One-To-Many: Use this mode when you '
                                                  'have got multiple target/S files',
                                      epilog=""""--> Sample Usage: watermark otm -t 'this 
                                              is my watermark text' -i 'sample1.pdf' 'sample2.pdf' 
                                              'MyPC/Downloads/PDFFiles'""")
        otm.add_argument('-t', '--text', type=str, default='PDFPlay', dest='text',
                         required=True, action=actions.watermark_text, nargs='+',
                         help='(**) Text that is to be applied as the watermark')
        otm.add_argument('-i', '--input', default=None, type=types.target_file_otm,
                         action=actions.target_file_otm, dest='target_files', nargs='+',
                         required=True,
                         help="""(**) Paths to files/directories that contain
                                 your target files""")
        otm.add_argument('-o', '--output', default=None,
                         type=types.output_file_otm,
                         action=actions.output_file_otm, dest='output_directory',
                         help='Directory to save watermarked files')
        otm.add_argument('-fn', '--font-name', default='Helvetica-Bold', type=str,
                         choices=FONTS, dest='font_name',
                         help='Name of the font that you want to use in the watermark')
        otm.add_argument('-fs', '--font-size', default='medium', type=str,
                         choices=FONT_SIZES, dest='font_size',
                         help='Size of the font')
        otm.add_argument('-fc', '--font-color', default='lightred', type=str,
                         choices=COLORS, dest='font_color',
                         help='Color of the font')
        otm.add_argument('-ta', '--text-alignment', default='diagonal', type=str,
                         choices=['horizontal', 'diagonal'], dest='text_alignment',
                         help='Alignment of the watermark in the document')
        otm.add_argument('-p', '--print', action='store_true',
                         default=False, dest='verbose',
                         help='Display informational messages')

        mto = sub_commands.add_parser('mto', help='Many-To-One: Apply many different '
                                                  'watermarks to a single file',
                                      description="""Many-To-One: Use this mode when you
                                              want to watermark a file multiple times with
                                              different watermark texts""",
                                      epilog="""--> Sample usage: watermark mto -t 
                                              'watermark text#1' 'watermark text#2' 
                                              'MyPC/SampleWatermarks/sample1.txt' -i 
                                              'sample.pdf'""")
        mto.add_argument('-t', '--text', default='PDFPlay',
                         type=types.watermark_mto,
                         action=actions.watermark_mto,
                         dest='texts', nargs='+',
                         help="""(**) Texts that are to be watermarked or path to a txt
                                      file that contains such texts""")
        mto.add_argument('-i', '--input', default=None, type=types.target_file_mto,
                         action=actions.target_file_oto, dest='target_file',
                         required=True,
                         help='(**) Path to the PDF file that is to be watermarked')
        mto.add_argument('-o', '--output', default=None,
                         type=types.output_file_otm,
                         action=actions.output_file_otm, dest='output_directory',
                         help='Directory to save watermarked files')
        mto.add_argument('-fn', '--font-name', default='Helvetica-Bold', type=str,
                         choices=FONTS, dest='font_name',
                         help='Name of the font that you want to use in the watermark')
        mto.add_argument('-fs', '--font-size', default='medium', type=str,
                         choices=FONT_SIZES, dest='font_size',
                         help='Size of the font')
        mto.add_argument('-fc', '--font-color', default='lightred', type=str,
                         choices=COLORS, dest='font_color',
                         help='Color of the font')
        mto.add_argument('-ta', '--text-alignment', default='diagonal', type=str,
                         choices=['horizontal', 'diagonal'], dest='text_alignment',
                         help='Alignment of the watermark in the document')
        mto.add_argument('-p', '--print', action='store_true',
                         default=False, dest='verbose',
                         help='Display informational messages')

        args = parser.parse_args()
        args.command = 'watermark'
        _validate_args(parser, args)
        args = _update_args(args)
        return args
    except Exception as e:
        print(f'Error --> Bad user-input: {e}')
        raise e


def _test():
    command = ['watermark', '-t', 'test target']
    args = parse_user_args(command)
    print(f'User Args: {args}')


if __name__ == '__main__':
    _test()



from argparse import Action
from os import listdir
from os import makedirs
from os.path import isdir
from os.path import isfile
from os.path import join
from os.path import sep
from pathlib import Path
import pandas as pd

from pdf_play.helpers import utils


class _WatermarkText(Action):

    def __call__(self, parser, namespace, values, option_string=None):
        setattr(namespace, self.dest, ' '.join(values))


class _TargetFileOTO(Action):

    def __call__(self, parser, namespace, values, option_string=None):
        setattr(namespace, self.dest, values)


class _OutputFileOTO(Action):

    def __call__(self, parser, namespace, values, option_string=None):
        base = Path(values).parent.resolve()
        if not isdir(base):
            makedirs(base)
        if not values.endswith('.pdf'):
            values += '.pdf'
        setattr(namespace, self.dest, values)


class _TargetFileOTM(Action):

    def __call__(self, parser, namespace, values, option_string=None):
        values_ = []
        for value in values:
            if isdir(value):
                pdfs = list(map(lambda x: join(value, x), filter(utils.is_pdf, listdir(
                    value))))
                values_.extend(pdfs)
            else:
                values_.append(value)
        values = list(set(values_))
        setattr(namespace, self.dest, values)


class _OutputFileOTM(Action):

    def __call__(self, parser, namespace, values, option_string=None):
        base = Path(values).parent.resolve()
        target = values.split(sep)[-1]
        existing_dirs = len(list(filter(lambda x: target in x, listdir(base))))
        if existing_dirs:
            values = join(base, f'{target}({existing_dirs + 1})')
        else:
            values = join(base, target)
        if not isdir(values):
            makedirs(values)
        setattr(namespace, self.dest, values)


class _WatermarkMTO(Action):

    def __call__(self, parser, namespace, values, option_string=None):
        values_ = []
        for value in values:
            if isfile(value):
                if value.endswith('.csv'):
                    data = pd.read_csv(value)
                else:
                    data = pd.read_excel(value)
                values_.extend(data.iloc[:, :2].values.tolist())
                continue
            values_.append(value)
        setattr(namespace, self.dest, values_)


target_file_oto = _TargetFileOTO
output_file_oto = _OutputFileOTO
target_file_otm = _TargetFileOTM
output_file_otm = _OutputFileOTM
watermark_mto = _WatermarkMTO
watermark_text = _WatermarkText



from io import BytesIO
from math import atan
from math import degrees
from math import sqrt

from PyPDF4.pdf import PdfFileReader
from reportlab.lib.colors import Color
from reportlab.pdfgen.canvas import Canvas
from pdf_play.core import __settings__ as settings


# noinspection PyTypeChecker
class Watermark:

    def __init__(self, text, page_size='A4', font_name='Helvetica-Bold',
                 font_size='medium',
                 text_alignment='diagonal', font_color='black', position_x='center',
                 position_y='center', verbose=False, debug=False):
        # TODO: assert value check
        # TODO: cleanup
        self._page_size = page_size
        self._font_name = font_name
        self.font_size = font_size
        self._font_size = 50
        self._text_alignment = text_alignment
        self._color = font_color
        self._text = text
        self._position_x = position_x
        self._position_y = position_y
        self.verbose = verbose
        self.debug = debug
        self._x, self._y = self._page_size
        self._d = int(sqrt(self._x ** 2 + self._y ** 2))
        self._init_canvas()
        self._update_style()
        # self._update_canvas()

    def _init_canvas(self):
        self._buffer = BytesIO()
        self._canvas = Canvas(self._buffer)

    def _set_rotation(self):
        x, y = list(map(float, self._page_size))
        rotation = 0
        if self._text_alignment == 'diagonal':
            rotation = round(degrees(atan(y / x)))
        if self._text_alignment == 'vertical':
            rotation = 90
        self._rotation = rotation

    def _set_max_length(self):
        """
            80% of axis length
            depends on text alignment:
                - horizontal: x-axis length
                - vertical: y-axis length
                - diagonal: diagonal length
        """
        lengths = dict(horizontal=self._x, vertical=self._y, diagonal=self._d)
        length = lengths[self._text_alignment]
        self._max_length = int(length * .8)

    def _set_font_size(self):
        x, y = self._page_size
        max_width = round(self._max_length * .5)
        max_font = x * .1
        if self.font_size == 'medium':
            max_width = round(self._max_length * .75)
            max_font = x * .2
        if self.font_size == 'large':
            max_width = round(self._max_length)
            max_font = x * .3
        size = 5

        while True:
            width = self._canvas.stringWidth(self._text,
                                             fontName=self._font_name,
                                             fontSize=size)
            if width >= max_width or size >= max_font:
                break
            size += 5
        self._font_size = size
        self._width = width

    def _set_color(self):
        color = settings.COLORS[self._color]
        self._red = color['red'] / 255
        self._green = color['green'] / 255
        self._blue = color['blue'] / 255
        self._alpha = color['alpha']

    def _set_position(self):
        # set to origin
        self._x = 0
        self._y = 0

    def _update_canvas(self):
        # update canvas attrs
        self._canvas.setPageSize(self._page_size)
        self._canvas.setFont(self._font_name, self._font_size)
        self._canvas.setFillColor((self._red, self._green, self._blue), alpha=self._alpha)
        # TODO: object attrs
        x, y = list(map(int, self._page_size))
        cx, cy = (x // 2), (y // 2)  # - (self._font_size * 2)
        # set origin to center of the page
        self._canvas.translate(cx, cy)
        self._canvas.rotate(self._rotation)
        x, y = (0, 0)
        self._canvas.drawCentredString(x, y,
                                       'a really long email id @ gmail.com but it'
                                       ' needs to be longer')
        # self._canvas.drawCentredString(self._x, -.25 * self._font_size,
        #                                self._text)
        # self._canvas.save()

    def _update_style(self):
        self._set_max_length()
        self._set_rotation()
        self._set_font_size()
        self._set_color()
        # self._set_position()

    def _calculate_font_size(self, text=None, font=None):
        if text is None:
            text = self._text
        if font is None:
            font = self._font_name
        factors = dict(small=.3, medium=.5, large=.7)
        max_width = int(self._max_length * factors[self.font_size])
        size = 2
        while True:
            width = self._canvas.stringWidth(text, font, size)
            if width > max_width or size > max_width*.2:
                break
            size += 2
        return size

    def _draw_watermark(self):
        """
            update canvas style
            draw watermark text
        """
        texts = self._text.splitlines()
        self._canvas.setPageSize(self._page_size)
        self._canvas.setFont(self._font_name, self._font_size)
        self._canvas.setFillColor((self._red, self._green, self._blue), alpha=self._alpha)
        # set origin to center of the page
        self._canvas.translate(self._page_size[0] // 2, self._page_size[1] // 2)
        self._canvas.rotate(self._rotation)
        font_size = self._calculate_font_size(max(texts, key=len))
        x, y = (0, (len(texts)//2) * (font_size*.75))
        for i, text in enumerate(texts):
            self._canvas.setFontSize(font_size)
            self._canvas.drawCentredString(x, y, text)
            y -= (font_size * .8) + 10
        self._canvas.save()

    def unload(self):
        self._draw_watermark()
        watermarked_pdf = PdfFileReader(self._buffer)
        return watermarked_pdf.getPage(0)



from pdf_play.core._pdf import PDF

client = PDF(debug=True)


def watermark(wm_text, file_to_watermark, file_to_save_it_as, **style):
    print(f'watermarking: {locals()}')
    client.apply_watermark(wm_text, file_to_watermark, file_to_save_it_as, **style)
    return True


__all__ = ['watermark']
